/* $Id: xpci_adapter.c,v 1.1 2004/03/04 23:04:03 moleres Exp $ */
/******************************************************************************
*
*       XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS"
*       AS A COURTESY TO YOU, SOLELY FOR USE IN DEVELOPING PROGRAMS AND
*       SOLUTIONS FOR XILINX DEVICES.  BY PROVIDING THIS DESIGN, CODE,
*       OR INFORMATION AS ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE,
*       APPLICATION OR STANDARD, XILINX IS MAKING NO REPRESENTATION
*       THAT THIS IMPLEMENTATION IS FREE FROM ANY CLAIMS OF INFRINGEMENT,
*       AND YOU ARE RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE
*       FOR YOUR IMPLEMENTATION.  XILINX EXPRESSLY DISCLAIMS ANY
*       WARRANTY WHATSOEVER WITH RESPECT TO THE ADEQUACY OF THE
*       IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR
*       REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE FROM CLAIMS OF
*       INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
*       FOR A PARTICULAR PURPOSE.
*
*       (c) Copyright 2004 Xilinx Inc.
*       All rights reserved.
*
******************************************************************************/
/*****************************************************************************/
/**
*
* @file xpci_adapter.c
*
* The implementation of the XPciAdapter component, which is the adapter
* between VxWorks and the XPci driver. See the header file, xpci_adapter.h,
* for more information.
*
* <pre>
* MODIFICATION HISTORY:
*
* Ver   Who  Date     Changes
* ----- ---- -------- -----------------------------------------------
* 1.00a rpm  02/27/04 First release
* </pre>
*
*****************************************************************************/

/***************************** Include Files ********************************/

#include "xparameters.h"
#include "xbasic_types.h"
#include "xstatus.h"
#include "xpci_adapter.h"

/************************** Constant Definitions ****************************/


/**************************** Type Definitions ******************************/

/*
 * This typedef contains configuration information for the device.
 */
typedef struct
{
    Xuint16 DeviceId;       /* ID to identify the XPci driver */
    int Used;               /* Indicate whether this entry is used */
    XPci Pci;               /* XPci driver instance */
} PciDriverInfo;


/***************** Macros (Inline Functions) Definitions ********************/


/************************** Function Prototypes *****************************/

/* Function to lookup driver info based on device ID */
static PciDriverInfo *LookupDriver(Xuint16 DeviceId);

/************************** Variable Definitions ****************************/

/* Maintain an array of XPci driver instance data */
static PciDriverInfo PciDriver[XPAR_XPCI_NUM_INSTANCES];

/****************************************************************************/
/**
*
* Initialize the adapter and the Xilinx device driver. Upon a successful return
* from this function, the XPci device driver instance (for the corresponding
* unique device ID) is initialized and can be used directly by an application
* or BSP.
*
* @param DeviceId is the unique device ID of the PCI device being initialized.
* @param BusNo is the PCI bus number of the Xilinx PCI device.
* @param SubBusNo is the PCI sub-bus number of the Xilinx PCI device.
*
* @return
*
* XST_SUCCESS on successful initialization, or a driver-specific error
* code on failure.
*
* @note
*
* None.
*
*****************************************************************************/
XStatus XPciAdapter_Init(int DeviceId, int BusNo, int SubBusNo)
{
    XStatus Result;
    PciDriverInfo *DriverInfo;

    DriverInfo = LookupDriver((Xuint16)DeviceId);
    if (DriverInfo == XNULL)
    {
        /* Too many PCI devices have already been initialized */
        return XST_FAILURE; /* TODO: better return value? */
    }

    Result = XPci_Initialize(&DriverInfo->Pci, DeviceId, BusNo, SubBusNo);
    if (Result != XST_SUCCESS)
    {
        return Result;
    }

    return XST_SUCCESS;
}


/****************************************************************************/
/**
*
* Get a pointer to the Xilinx device driver instance for the specified devie
* ID. This pointer can be used to access the Xilinx PCI device driver API
* (see xpci.h).
*
* @param DeviceId is the unique device ID of the PCI device being initialized.
*
* @return
*
* A pointer to the XPci driver instance data of the specified device, or XNULL
* if no matching device is found. Note that the XPci driver is not guaranteed
* to be initialized. The user must call XPciAdapter_Init() successfully before
* calling this function to guarantee the driver is initialized.
*
* @note
*
* None.
*
*****************************************************************************/
XPci *XPciAdapter_GetDriver(int DeviceId)
{
    PciDriverInfo *DriverInfo;

    DriverInfo = LookupDriver((Xuint16)DeviceId);
    if (DriverInfo == XNULL)
    {
        return XNULL;
    }

    return &DriverInfo->Pci;
}

/****************************************************************************/
/**
*
* Lookup the driver based on device ID.  The static array of PCI drivers is
* searched and either an entry that matches the specified device ID is
* returned, or the first unused entry is returned.  If all entries are used,
* a value of XNULL is return.
*
* @param DeviceId is the unique device ID of the PCI device being sought.
*
* @return
*
* A pointer to the adapter's PciDriverInfo entry for the specified device, or
* an unused PciDriverInfo entry, or XNULL if all entries are already used.
*
* @note
*
* None.
*
*****************************************************************************/
static PciDriverInfo *LookupDriver(Xuint16 DeviceId)
{
    int i;
    int UnusedIndex = -1;       /* first unused entry */

    for (i=0; i < XPAR_XPCI_NUM_INSTANCES; i++)
    {
        if (PciDriver[i].DeviceId == DeviceId)
        {
            /*
             * Found a match, return a pointer to the found structure.
             */
            PciDriver[i].Used = 1;
            return &PciDriver[i];
        }

        /*
         * Keep track of the first entry in the table that is unused
         */
        if ((UnusedIndex == -1) && (PciDriver[i].Used == 0))
        {
            UnusedIndex = i;
        }
    }

    /*
     * No match was found, so if there is an unused table entry, use it
     * and store the input device ID.
     */
    if (UnusedIndex != -1)
    {
        PciDriver[i].DeviceId = DeviceId;
        PciDriver[i].Used = 1;
        return &PciDriver[UnusedIndex];
    }

    /*
     * All table entries are currently used, so return null
     */
    return XNULL;
}

