/* $Id: xsysaceblkadapter.c,v 1.4 2003/07/24 16:37:11 moleres Exp $ */
/******************************************************************************
*
*       XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS"
*       AS A COURTESY TO YOU, SOLELY FOR USE IN DEVELOPING PROGRAMS AND
*       SOLUTIONS FOR XILINX DEVICES.  BY PROVIDING THIS DESIGN, CODE,
*       OR INFORMATION AS ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE,
*       APPLICATION OR STANDARD, XILINX IS MAKING NO REPRESENTATION
*       THAT THIS IMPLEMENTATION IS FREE FROM ANY CLAIMS OF INFRINGEMENT,
*       AND YOU ARE RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE
*       FOR YOUR IMPLEMENTATION.  XILINX EXPRESSLY DISCLAIMS ANY
*       WARRANTY WHATSOEVER WITH RESPECT TO THE ADEQUACY OF THE
*       IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR
*       REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE FROM CLAIMS OF
*       INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
*       FOR A PARTICULAR PURPOSE.
*
*       (c) Copyright 2002 Xilinx Inc.
*       All rights reserved.
*
******************************************************************************/
/*****************************************************************************/
/**
*
* @file xsysaceblkadapter.c
*
* The Xilinx System ACE Block Adapter implementation. See xsysace_blk_adapter.h
* for a detailed description of this adapter.
*
* <pre>
* MODIFICATION HISTORY:
*
* Ver   Who  Date     Changes
* ----- ---- -------- -----------------------------------------------
* 1.00a rpm  07/31/02 First release
* </pre>
*
******************************************************************************/

/***************************** Include Files *********************************/
#include "xbasic_types.h"
#include "xstatus.h"            /* status codes */
#include "xparameters.h"
#include "xsysaceblkadapter.h"
#include "xsysace.h"

#include "semLib.h"
#include "errnoLib.h"       /* for errno */
#include "stdlib.h"         /* for calloc */
#include "ioLib.h"          /* for O_RDWR and ioctl commands */

/************************** Constant Definitions *****************************/

/**************************** Type Definitions *******************************/

/*
 * A typedef for a partition, or logical disk. Every CompactFlash storage
 * device may have one or more partitions on it.
 */
typedef struct
{
    BLK_DEV BlkDev;              /* Must be first element of structure */
    unsigned int Id;             /* ID (index) of this partition */
    unsigned int Offset;         /* Offset, in blocks, of partition */
    unsigned int NumBlks;        /* Number of blocks in this partition */
    struct AceInfoTag *InfoPtr;  /* Convenient pointer to driver info */

} Partition;


/*
 * A typedef for each System ACE/CompactFlash device. Each System ACE/CF
 * has a partition array (one or more logical disks).
 */
typedef struct AceInfoTag
{
    Partition *PartitionPtr;        /* Partition array */
    Xuint16 DeviceId;               /* Unique device ID of ACE controller */
    XSysAce Ace;                    /* Driver component */
    XSysAce_CFParameters Params;    /* CompactFlash parameters */
    unsigned int NumPartitions;     /* Number of partitions in the CF */
    SEM_ID MutexSem;                /* Binary semaphore for mutual exclusion */

} AceInfo;


/***************** Macros (Inline Functions) Definitions *********************/


/************************** Function Prototypes ******************************/

/*
 * The VxWorks block device interface functions. These can be static since
 * they'll be part of a function table passed to VxWorks in the BLK_DEV
 * structure.
 */
static STATUS BlkRead(Partition *PartPtr, int StartBlk, int NumBlks,
                      char *BufferPtr);
static STATUS BlkWrite(Partition *PartPtr, int StartBlk, int NumBlks,
                       char *BufferPtr);
static STATUS Ioctl(Partition *PartPtr, int Cmd, int Arg);
static STATUS StatusCheck(Partition *PartPtr);

/*
 * Internal functions
 */
static Xboolean IsValidPartition(AceInfo *InfoPtr, unsigned PartitionId,
                                 unsigned NumBlks, unsigned Offset);
static AceInfo *LookupDevice(Xuint16 DeviceId);
static void Cleanup(AceInfo *InfoPtr);
static STATUS IoctlFormat(Partition *PartPtr);

/************************** Variable Definitions *****************************/

/*
 * Each ACE/CompactFlash device gets its own structure which helps the
 * adapter keep track of all device-related information.
 */
static AceInfo AceDevice[XPAR_XSYSACE_NUM_INSTANCES];


/*****************************************************************************/
/**
*
* Initialize the System ACE controller corresponding to the given device ID.
* This function should be called only once for each ACE controller in the
* system.  Allocates a Partition structure for every partition possible on
* the CompactFlash storage device (per the given argument).
*
* The CompactFlash identify parameters are retrieved during initialization
* since they are used elsewhere within the adapter. Since a read of the
* CompactFlash device is required, a binary semphore is used to prevent other
* tasks from reading or writing this same CompactFlash storage device - since
* only one reader/writer can be active at a time.
*
* @param DeviceId        The unique device ID of the ACE controller being
*                        initialized.
* @param MaxPartitions   The maximum number of partitions that are to be
*                        created on the CompactFlash storage device.
*
* @return
*
* OK if successful, or ERROR if an error occurs. If an error occurs, errno
* will be set appropriately.
*
* @note
*
* This function blocks waiting on a lock from the System ACE controller. If
* a JTAG configuration is in process, this function will block until the
* configuration process is done.
*
******************************************************************************/
STATUS XSysAceBlkAdapter_Init(Xuint16 DeviceId, unsigned MaxPartitions)
{
    XStatus Result;
    AceInfo *InfoPtr;

    /*
     * First get an AceInfo pointer for this device ID. If this device has
     * been previously initialized, we should get the same pointer.
     */
    InfoPtr = LookupDevice(DeviceId);
    if (InfoPtr == XNULL)
    {
        /* No more room based on the number of ACE controllers in the system */
        (void)errnoSet(ENODEV);
        return ERROR;
    }

    /*
     * Validate the number of partitions passed in by the user
     */
    if (MaxPartitions == 0)
    {
        (void)errnoSet(EINVAL);
        return ERROR;
    }

    /*
     * Set up the AceInfo structure
     */
    InfoPtr->DeviceId = DeviceId;
    InfoPtr->NumPartitions = MaxPartitions;

    /*
     * Dynamically allocate space for the number of partitions specified. For
     * each partition, we need enough space for the Partition structure.
     */
    InfoPtr->PartitionPtr = calloc(MaxPartitions, sizeof(Partition));
    if (InfoPtr->PartitionPtr == NULL)
    {
        /* not enough space for this */
        (void)errnoSet(ENOSPC);
        return ERROR;
    }

    /*
     * Create the mutual exclusion semaphore
     */
    InfoPtr->MutexSem = semBCreate(SEM_Q_FIFO, SEM_FULL);
    if (InfoPtr->MutexSem == NULL)
    {
        /* not enough space for this */
        (void)errnoSet(ENOSPC);
        return ERROR;
    }

    /*
     * Initialize the System ACE driver
     */
    Result = XSysAce_Initialize(&InfoPtr->Ace, DeviceId);
    if (Result != XST_SUCCESS)
    {
        /* Free allocated space */
        Cleanup(InfoPtr);

        (void)errnoSet(Result);
        return ERROR;
    }

    /*
     * Take the semaphore while we get the CompactFlash parameters
     */
    (void)semTake(InfoPtr->MutexSem, WAIT_FOREVER);

    /*
     * Get the CompactFlash parameters and use the number of LBA sectors as
     * the number of blocks on the device. We block waiting for a lock to
     * be granted.
     */
    while (XSysAce_Lock(&InfoPtr->Ace, XFALSE) != XST_SUCCESS);

    Result = XSysAce_IdentifyCF(&InfoPtr->Ace, &InfoPtr->Params);
    if (Result != XST_SUCCESS)
    {
        (void)errnoSet(Result);

        XSysAce_Unlock(&InfoPtr->Ace);
        (void)semGive(InfoPtr->MutexSem);

        /* Free allocated space */
        Cleanup(InfoPtr);
        return ERROR;
    }

    /* Release the lock and the semaphore */
    XSysAce_Unlock(&InfoPtr->Ace);
    (void)semGive(InfoPtr->MutexSem);

    return OK;
}


/*****************************************************************************/
/**
*
* Get the total number of blocks available on the CompactFlash storage
* device. This is a utility function for the BSP/application developer to
* determine the capacity of the storage device at run-time.
*
* @param DeviceId    The unique device ID of the System ACE controller being
*                    queried.
*
* @return
*
* The number of blocks supported by the CompactFlash storage device associated
* with the System ACE, or 0 if an error occurs. If an error occurs, errno
* will be set to indicate the cause of the error.
*
* @note
*
* None.
*
******************************************************************************/
unsigned int XSysAceBlkAdapter_GetTotalBlocks(Xuint16 DeviceId)
{
    AceInfo *InfoPtr;

    /*
     * Lookup the device object based on the given device ID
     */
    InfoPtr = LookupDevice(DeviceId);
    if (InfoPtr == NULL)
    {
        (void)errnoSet(ENODEV);
        return 0;
    }

    return InfoPtr->Params.LbaSectors;
}


/*****************************************************************************/
/**
*
* Create a block device for the given partition ID on the ACE/CompactFlash
* specified. We use the partition ID provided by the user as an index into
* the partition array that has already been created. The BLK_DEV structure of
* the partition is populated and returned to the caller. This function should
* be called once for each partition on a CompactFlash storage device.
*
* @param DeviceId       The unique device ID of the System ACE controller being
*                       worked on.
* @param PartitionId    The ID of the partition to be created. This ID
*                       ranges from 0 to one less than the maximum number of
*                       partitions passed to XSysAceBlkAdapter_Init().
* @param NumBlks        The number of blocks in this partition.
* @param Offset         The offset, or starting block number, for this
*                       partition. VxWorks always passes relative block numbers
*                       for reads and writes, so this offset will be added to
*                       that to get an absolute block number in the CompactFlash
*                       storage device.
*
* @return
*
* A pointer to the BLK_DEV structure for the corresponding partition ID, or
* NULL if an error occurs. If an error occurs, errno will be set to indicate
* the cause of the error. Although the return type is a BLK_DEV pointer, the
* value returned is actually a Partition pointer. Since a BLK_DEV is the first
* element of the Partition structure, VxWorks accesses it happily.  VxWorks
* passes this pointer to the block device functions defined in this adapter.
* These functions need the Partition pointer to get to the right block offset.
* This may sound kludgy, but this is how VxWorks intends it to be implemented.
*
* @note
*
* None.
*
******************************************************************************/
BLK_DEV *XSysAceBlkAdapter_DevCreate(Xuint16 DeviceId, unsigned PartitionId,
                                     unsigned NumBlks, unsigned Offset)
{
    AceInfo *InfoPtr;
    Partition *PartPtr;

    /*
     * Lookup the device object based on the given device ID
     */
    InfoPtr = LookupDevice(DeviceId);
    if (InfoPtr == NULL)
    {
        (void)errnoSet(ENODEV);
        return NULL;
    }

    /*
     * Validate the Partition ID and the number of blocks passed in by the user
     */
    if (!IsValidPartition(InfoPtr, PartitionId, NumBlks, Offset))
    {
        /* invalid partition parameters */
        (void)errnoSet(EINVAL);
        return NULL;
    }

    /*
     * Get a pointer to the correct Partition and populate it
     */
    PartPtr = InfoPtr->PartitionPtr + PartitionId;

    PartPtr->Id = PartitionId;
    PartPtr->NumBlks = NumBlks;
    PartPtr->Offset = Offset;
    PartPtr->InfoPtr = InfoPtr;

    /*
     * Populate the BLK_DEV structure. Note that there is no Reset function
     * as we cannot reset the CompactFlash or System ACE controller without
     * resetting all "partitions".  Also, this reset may cause an FPGA
     * reconfiguration, depending on how the System ACE is configured.
     */
    PartPtr->BlkDev.bd_blkRd = BlkRead;
    PartPtr->BlkDev.bd_blkWrt = BlkWrite;
    PartPtr->BlkDev.bd_ioctl = Ioctl;
    PartPtr->BlkDev.bd_reset = NULL;
    PartPtr->BlkDev.bd_statusChk = StatusCheck;
    PartPtr->BlkDev.bd_removable = TRUE;
    PartPtr->BlkDev.bd_nBlocks = NumBlks;
    PartPtr->BlkDev.bd_bytesPerBlk = XSA_CF_SECTOR_SIZE;
    PartPtr->BlkDev.bd_blksPerTrack = InfoPtr->Params.CurSectorsPerTrack;
    PartPtr->BlkDev.bd_nHeads = InfoPtr->Params.CurNumHeads;
    PartPtr->BlkDev.bd_retry = 2;       /* number of times to retry on error */
    PartPtr->BlkDev.bd_mode = O_RDWR;

    /* Always assume the CompactFlash is detected TODO: OK for now? */
    PartPtr->BlkDev.bd_readyChanged = TRUE;

    return (BLK_DEV *)PartPtr;
}

/*****************************************************************************/
/**
*
* Get a pointer to the XSysAce driver used for the given device ID. This call
* is only valid after the XSysAceBlkAdapter_Init() function has been called.
*
* @param DeviceId    The unique device ID of the System ACE controller being
*                    queried.
*
* @return
*
* A pointer to the XSysAce driver used for this device ID, or NULL if the
* device ID has not yet been initialized.
*
* @note
*
* None.
*
******************************************************************************/
XSysAce *XSysAceBlkAdapter_GetDriver(Xuint16 DeviceId)
{
    AceInfo *InfoPtr;

    /*
     * First get an AceInfo pointer for this device ID. If this device has
     * been previously initialized, we should get the same pointer.
     */
    InfoPtr = LookupDevice(DeviceId);
    if (InfoPtr == XNULL)
    {
        /*
         * No entry for this device ID and no more room based on the number of
         * ACE controllers in the system
         */
        (void)errnoSet(ENODEV);
        return NULL;
    }

    if (InfoPtr->PartitionPtr == NULL)
    {
        /*
         * This was just an empty entry, so return null since no block
         * adapter with this device ID has been initialized yet.
         */
        (void)errnoSet(ENODEV);
        return NULL;
    }

    return &InfoPtr->Ace;
}

/*****************************************************************************/
/**
*
* Perform a block read of the CompactFlash storage device. Note that this is
* a polled read of the ACE/CompactFlash device. There is no mechanism in the
* VxWorks block device interface for a read in interrupt mode.
*
* A binary semphore is used to prevent other tasks from reading or writing
* this same CompactFlash storage device - since only one reader/writer can be
* active at a time.
*
* @param PartPtr     A pointer to the partition (and BLK_DEV) from which
*                    to read. VxWorks uses the pointer returned from
*                    XSysAceBlkAdapter_DevCreate().
* @param StartBlk    The starting block from which to read.
* @param NumBlks     The number of blocks to read.
* @param BufferPtr   A pointer to a buffer where the data will be stored. This
*                    buffer is assumed to be big enough to hold the requested
*                    number of blocks.
*
* @return
*
* OK if successful, or ERROR if an error occurs. The global variable errno
* is set if an error occurs.
*
* @note
*
* This function blocks waiting on a lock from the System ACE controller. If
* a JTAG configuration is in process, this function will block until the
* configuration process is done. TODO: could return error instead?
*
******************************************************************************/
static STATUS BlkRead(Partition *PartPtr, int StartBlk, int NumBlks,
                      char *BufferPtr)
{
    XStatus Result;
    XSysAce *AcePtr = &PartPtr->InfoPtr->Ace;

    /*
     * Take the semaphore while we get the CompactFlash parameters
     */
    (void)semTake(PartPtr->InfoPtr->MutexSem, WAIT_FOREVER);

    /*
     * Block waiting for a lock to be granted.
     */
    while (XSysAce_Lock(AcePtr, XFALSE) != XST_SUCCESS);

    /*
     * Read the sectors as requested. Remember to add the offset of the
     * partition to the starting block.
     */
    Result = XSysAce_SectorRead(AcePtr, StartBlk + PartPtr->Offset,
                                NumBlks, BufferPtr);
    if (Result != XST_SUCCESS)
    {
        (void)errnoSet(Result);

        XSysAce_Unlock(AcePtr);
        (void)semGive(PartPtr->InfoPtr->MutexSem);
        return ERROR;
    }

    /* Wait for CF to be ready again, meaning read is really complete */
    while (!XSysAce_IsCFReady(AcePtr));

    /* Release the lock and the semaphore */
    XSysAce_Unlock(AcePtr);
    (void)semGive(PartPtr->InfoPtr->MutexSem);

    return OK;
}


/*****************************************************************************/
/**
*
* Perform a block write to the CompactFlash storage device. Note that this is
* a polled write of the ACE/CompactFlash device. There is no mechanism in the
* VxWorks block device interface for a write in interrupt mode.
*
* A binary semphore is used to prevent other tasks from reading or writing
* this same CompactFlash storage device - since only one reader/writer can be
* active at a time.
*
* @param PartPtr     A pointer to the partition (and BLK_DEV) to write.
*                    VxWorks uses the pointer returned from
*                    XSysAceBlkAdapter_DevCreate().
* @param StartBlk    The starting block where the write will start.
* @param NumBlks     The number of blocks to write.
* @param BufferPtr   A pointer to a buffer to be written.
*
* @return
*
* OK if successful, or ERROR if an error occurs. The global variable errno
* is set if an error occurs.
*
* @note
*
* This function blocks waiting on a lock from the System ACE controller. If
* a JTAG configuration is in process, this function will block until the
* configuration process is done. TODO: could return error instead?
*
******************************************************************************/
static STATUS BlkWrite(Partition *PartPtr, int StartBlk, int NumBlks,
                       char *BufferPtr)
{
    XStatus Result;
    XSysAce *AcePtr = &PartPtr->InfoPtr->Ace;

    /*
     * Take the semaphore while we get the CompactFlash parameters
     */
    (void)semTake(PartPtr->InfoPtr->MutexSem, WAIT_FOREVER);

    /*
     * Block waiting for a lock to be granted.
     */
    while (XSysAce_Lock(AcePtr, XFALSE) != XST_SUCCESS);

    /*
     * Write the sectors as requested. Remember to add the offset of the
     * partition to the starting block.
     */
    Result = XSysAce_SectorWrite(AcePtr, StartBlk + PartPtr->Offset,
                                 NumBlks, BufferPtr);
    if (Result != XST_SUCCESS)
    {
        (void)errnoSet(Result);

        XSysAce_Unlock(AcePtr);
        (void)semGive(PartPtr->InfoPtr->MutexSem);
        return ERROR;
    }

    /* Wait for CF to be ready again, meaning write is really complete */
    while (!XSysAce_IsCFReady(AcePtr));

    /* Release the lock and the semaphore */
    XSysAce_Unlock(AcePtr);
    (void)semGive(PartPtr->InfoPtr->MutexSem);

    return OK;
}


/*****************************************************************************/
/**
*
* Handle ioctl requests for the block device. The only ioctl command currently
* supported is FIODISKFORMAT. Other ioctl commands are supported by the
* filesystem on top of this adapter.
*
* @param PartPtr     A pointer to the partition (and BLK_DEV) to work with.
*                    VxWorks uses the pointer returned from
*                    XSysAceBlkAdapter_DevCreate().
* @param Cmd         The ioctl command.
* @param Arg         An argument to the ioctl command.
*
* @return
*
* OK if successful, or ERROR if an error occurs. The global variable errno
* is set to S_ioLib_UNKNOWN_REQUEST if the command is not a recognized command.
*
* @note
*
* None.
*
******************************************************************************/
static STATUS Ioctl(Partition *PartPtr, int Cmd, int Arg)
{
    STATUS Status = ERROR;

    if (Cmd == FIODISKFORMAT)
    {
        Status = IoctlFormat(PartPtr);
    }
    else
    {
        (void)errnoSet(S_ioLib_UNKNOWN_REQUEST);
    }

    return Status;
}


/*****************************************************************************/
/**
*
* Perform a status check on the partition. This routine detects the entire
* CompactFlash storage device and verifies that is it ready. There is no
* per-partition status check.
*
* @param PartPtr     A pointer to the partition (and BLK_DEV) to check.
*                    VxWorks uses the pointer returned from
*                    XSysAceBlkAdapter_DevCreate().
*
* @return
*
* OK if successful, meaning the it is OK to open or create files on the
* partition.  Return ERROR if an error occurs. The global variable errno
* is set if an error occurs.
*
* @note
*
* None.
*
******************************************************************************/
static STATUS StatusCheck(Partition *PartPtr)
{
    XSysAce *AcePtr = &PartPtr->InfoPtr->Ace;

    if (XSysAce_IsCFReady(AcePtr))
    {
        return OK;
    }

    return ERROR;
}


/*****************************************************************************/
/**
*
* Look in the AceDevice table for an existing device that matches the given
* DeviceId, or an empty spot in the table if no match is found. This function
* is used by every public function defined above that takes a DeviceId as
* argument.
*
* @param DeviceId   The devide ID to search for in the AceDevice table
*
* @return
*
* Returns a pointer to an existing AceInfo element if a match is found, or
* an unused AceDevice element if no match is found. Returns XNULL if no match
* is found and the AceDevice array is full (i.e., all elements have been used).
*
* NOTES:
*
* Assumes that this would never be called with a Device ID of zero!!!!!
*
******************************************************************************/
static AceInfo *LookupDevice(Xuint16 DeviceId)
{
    int i;
    int UnusedIndex = -1;       /* first unused entry */

    for (i=0; i < XPAR_XSYSACE_NUM_INSTANCES; i++)
    {
        if (DeviceId == AceDevice[i].DeviceId)
        {
            /* Found a match */
            return &AceDevice[i];
        }

        /*
         * Keep track of the first entry in the table that is unused. We
         * base the unused on the number of partitions set for the device.
         */
        if ((UnusedIndex == -1) && (AceDevice[i].NumPartitions == 0))
        {
            UnusedIndex = i;
        }
    }

    /*
     * No match was found, so if there is an unused table entry, use it,
     */
    if (UnusedIndex != -1)
    {
        return &AceDevice[UnusedIndex];
    }

    /*
     * All table entries are currently used, so return null
     */
    return XNULL;
}

/*****************************************************************************/
/**
*
* Free any dynamically allocated memory.
*
* @param InfoPtr     A pointer to the AceInfo structure being worked on.
*
* @return
*
* None.
*
* @note
*
* None.
*
******************************************************************************/
static void Cleanup(AceInfo *InfoPtr)
{
    /*
     * Free the partition if we allocated it
     */
    if (InfoPtr->PartitionPtr != XNULL)
    {
        free(InfoPtr->PartitionPtr);
        InfoPtr->PartitionPtr = XNULL;
    }

    /*
     * Delete the mutex semaphore. First take it to be sure we do
     * not delete while another task has it.
     */
    if (InfoPtr->MutexSem != XNULL)
    {
        (void)semTake(InfoPtr->MutexSem, WAIT_FOREVER);
        (void)semDelete(InfoPtr->MutexSem);
        InfoPtr->MutexSem = XNULL;
    }
}


/*****************************************************************************/
/**
*
* Format the disk as a result of an ioctl command. The disk is formatted by
* zeroing out each sector, one at a time.  Multiple sectors, up to 256, can
* be written to in a single call, but this did not help performance. The
* format operation is a slow one - and remember we're in polled mode.
*
* @param PartPtr     A pointer to the partition being worked on.
*
* @return
*
* OK if successful, or ERROR if an error occurs.
*
* @note
*
* None.
*
******************************************************************************/
static STATUS IoctlFormat(Partition *PartPtr)
{
    XStatus Result;
    STATUS Status = ERROR;
    AceInfo *InfoPtr = PartPtr->InfoPtr;
    char *ZeroBuffer;

    /*
     * Write zeros to every sector of the partition. Since sector sizes
     * may vary, we dynamically allocate zero-filled buffer space for
     * one sector, then write it over and over to disk.
     */
    ZeroBuffer = calloc(InfoPtr->Params.NumBytesPerSector, 1);
    if (ZeroBuffer != NULL)
    {
        int i;

        /*
         * Take the semaphore
         */
        (void)semTake(InfoPtr->MutexSem, WAIT_FOREVER);

        /*
         * Block waiting for a lock to be granted.
         */
        while (XSysAce_Lock(&InfoPtr->Ace, XFALSE) != XST_SUCCESS);

        Status = OK;        /* assume success for now */

        /*
         * Write the zero buffer to each sector
         */
        for (i = 0; i < PartPtr->NumBlks; i++)
        {
            /*
             * We need to wait until the CompactFlash is ready again before
             * trying to invoke another operation.
             */
            while (!XSysAce_IsCFReady(&InfoPtr->Ace));

            Result = XSysAce_SectorWrite(&InfoPtr->Ace, PartPtr->Offset + i,
                                         1, ZeroBuffer);
            if (Result != XST_SUCCESS)
            {
                (void)errnoSet(Result);
                Status = ERROR;
                break;
            }
        }

        /* Release the lock and the semaphore, free the space */
        XSysAce_Unlock(&InfoPtr->Ace);
        (void)semGive(InfoPtr->MutexSem);
        free(ZeroBuffer);
    }

    return Status;
}

/*****************************************************************************/
/**
*
* Validate the partition about to be created. Verify that the number of blocks
* and offset are in the correct range, and also verify that this partition
* does not overlap with another.
*
* @param InfoPtr        A pointer to the ACE being worked on.
* @param PartitionId    The ID of the partition to be created. This ID
*                       ranges from 0 to one less than the maximum number of
*                       partitions passed to XSysAceBlkAdapter_Init().
* @param NumBlks        The number of blocks in this partition.
* @param Offset         The offset, or starting block number, for this
*                       partition. VxWorks always passes relative block numbers
*                       for reads and writes, so this offset will be added to
*                       that to get an absolute block number in the CompactFlash
*                       storage device.
*
* @return
*
* XTRUE if it is a valid partition, XFALSE if it is not.
*
* @note
*
* None.
*
******************************************************************************/
static Xboolean IsValidPartition(AceInfo *InfoPtr, unsigned PartitionId,
                                 unsigned NumBlks, unsigned Offset)
{
    Partition *PartPtr;
    unsigned BlkStart;
    unsigned BlkEnd;
    int i;

    /*
     * Validate the Partition ID, which is an index into the partition array
     */
    if (PartitionId >= InfoPtr->NumPartitions)
    {
        return XFALSE;
    }

    /*
     * Validate the offset is within range
     */
    if (Offset >= InfoPtr->Params.LbaSectors)
    {
        return XFALSE;
    }

    /*
     * Validate the number of blocks (make sure it fits into the CompactFlash)
     */
    if ((NumBlks == 0) || (NumBlks > (InfoPtr->Params.LbaSectors - Offset)))
    {
        return XFALSE;
    }

    /*
     * Now verify there is not an existing partition in the same space
     */
    BlkStart = Offset;
    BlkEnd = BlkStart + NumBlks;

    for (i=0; i < InfoPtr->NumPartitions; i++)
    {
        PartPtr = InfoPtr->PartitionPtr + i;

        /*
         * Do not check same partition and use InfoPtr to indicate whether
         * the partition has been initialized.
         */
        if ((PartPtr->Id != PartitionId) && (PartPtr->InfoPtr != NULL))
        {
            unsigned Start;
            unsigned End;

            Start = PartPtr->Offset;
            End = Start + PartPtr->NumBlks;

            /*
             * Check to see if the start or end of the partition is overlapped
             * with the partition we're validating.
             */
            if (((Start >= BlkStart) && (Start <= BlkEnd)) ||
                ((End >= BlkStart) && (End <= BlkEnd)))
            {
                return XFALSE;
            }
        }
    }

    return XTRUE;
}
